/* Copyright 2009 DJ Delorie <dj@delorie.com> */
/* Distributed under the terms of the GNU GPL, version 3. */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include "escp2.h"

static FILE *f;

static void
esc (const char *s, const char *fmt, ...)
{
  va_list v;
  int i;

  fputc (27, f);
  fputs (s, f);

  va_start (v, fmt);
  while (*fmt)
    {
      switch (*fmt++)
	{
	case 'b':
	  i = va_arg (v, int);
	  fputc (i % 256, f);
	  fputc (i / 256, f);
	  break;
	case 'i':
	  i = va_arg (v, int);
	  fputc ((i >> 0) & 0xff, f);
	  fputc ((i >> 8) & 0xff, f);
	  fputc ((i >> 16) & 0xff, f);
	  fputc ((i >> 24) & 0xff, f);
	  break;
	case 'c':
	  i = va_arg (v, int);
	  fputc (i % 256, f);
	  break;
	case 's':
	  s = va_arg (v, const char *);
	  fputs (s, f);
	  break;
	}
    }
}

static const char *output_file = "output.esc";

void
escp2_set_output (const char *name)
{
  output_file = name;
}

void
escp2_init (double page_width, double page_length, int draft_mode)
{
  f = fopen(output_file, "wb");

  fputc(0, f);
  fputc(0, f);
  fputc(0, f);
  fputc(0, f);
  fprintf(f, "\033\1@EJL 1284.4\n@EJL     \n\033@");

  esc("@", ""); /* reset */

  esc("(G", "bc", 1, 1); /* graphics mode */
  esc("(U", "bccccc", 5, 2, 2, 2, 128, 22); /* units 1/2880" */
  esc("(K", "bcc", 2, 0, 2); /* color mode */
  esc("(i", "bc", 1, 0); /* non-microweave */
  switch (draft_mode)
    {
    case 0: /* super accurate */
      esc("(s", "bc", 1, 2); /* speed 2 = 10"/s */
      esc("U", "c", 1); /* unidirectional */
      esc("(e", "bcc", 2, 0, 37); /* dotsize - smallest */
      break;
    case 1: /* plain paper accurate */
      esc("(s", "bc", 1, 0); /* speed 0 = 20"/s */
      esc("U", "c", 1); /* unidirectional */
      esc("(e", "bcc", 2, 0, 37); /* dotsize - smallest */
      break;
    case 2: /* fastest */
      esc("(s", "bc", 1, 0); /* speed 0 = 20"/s */
      esc("U", "c", 0); /* bidirectional */
      esc("(e", "bcc", 2, 0, 0); /* dotsize - biggest */
      break;
    }
  esc("(D", "bbcc", 4, 14400, 160, 20); /* res: 90 x 720 dpi */
  esc("(C", "bi", 4, (int)(page_length*2880));
  esc("(c", "bii", 8, 0, (int)(page_length*2880));
  esc("(S", "bii", 8, (int)(page_width*2880), (int)(page_length*2880));
  esc("(m", "bc", 1, 0xb0); /* print method (?) */
}

void
escp2_done ()
{
  fputc(12, f);
  esc("@", "");
  fclose (f);
}

void
escp2_advance_row (int vrel)
{
  esc("(v", "bi", 4, vrel);
}

void
escp2_hpos (int habs)
{
  esc("($", "bi", 4, habs);
}

static int bitw;
static unsigned char datum;
static int bitp;

static unsigned char dbuf[256];
int dbufp = 0;

void
escp2_band (int color, int bits, int width, int lines)
{
  int bytes = (bits*width+7)/8;
  esc("i", "cccbb", color, 1, bits, bytes, lines);
  bitw = bits;
  datum = 0;
  bitp = 0;
  dbufp = 0;
}

static int
rlen (int ofs)
{
  int i;
  for (i=ofs; i<dbufp; i++)
    if (dbuf[i] != dbuf[ofs])
      return i-ofs;
  return dbufp-ofs;
}

static void
flush_dbuf (int all)
{
  int i, r, u;
  while (1)
    {
      if (!all && dbufp < 128)
	break;
      if (dbufp == 0)
	break;
      u = 0;
      for (i=0; i<dbufp; i++)
	{
	  r = rlen (i);
	  if (r > 3)
	    break;
	}
      if (i)
	{
	  fputc (i-1, f);
	  fwrite (dbuf, i, 1, f);
	  u += i;
	}
      if (r > 3)
	{
	  fputc (257-r, f);
	  fputc (dbuf[u], f);
	  u += r;
	}
      memmove (dbuf, dbuf+u, dbufp-u);
      dbufp -= u;
    }
}

void
escp2_dot (int v)
{
  datum |= (v << (7-bitp));
  bitp += bitw;
  if (bitp >= 8)
    {
      dbuf[dbufp++] = datum;
      datum = 0;
      bitp = 0;
    }
  flush_dbuf (0);
}

void
escp2_end_row ()
{
  while (bitp)
    escp2_dot(0);
  flush_dbuf (1);
}

void
escp2_end_band ()
{
  fputc(13, f);
}
